@@ -1,176 +1,3 @@
-
-#' @title Get a clusterization running the `Seurat` package
-#'
-#' @description The function uses the [Seurat-package] to clusterize the given
-#'   counts raw data.
-#'
-#' @details The parameter resolution is set at 0.5 initially, but in case of too
-#'   few clusters it can be raised up to 2.5.
-#'
-#' @param rawData The raw counts
-#' @param cond The sample condition
-#' @param iter The current iteration
-#' @param initialResolution The resolution to use at first in the
-#'   *clusterization* algorithm
-#' @param minNumClusters The minimum number of *clusters* expected from this
-#'   *clusterization*. In cases it is not reached, it will increase the
-#'   resolution of the *clusterization*
-#' @param genesSel Decides whether and how to perform the gene-selection. used
-#'   for the clustering. It is a string indicating one of the following
-#'   selection methods:
-#'   * `"HGDI"` Will pick-up the genes with highest **GDI**
-#'   * `"HVG_Seurat"` Will pick-up the genes with the highest variability
-#'     via the \pkg{Seurat} package (the default method)
-#'   * `"HVG_Scanpy"` Will pick-up the genes with the highest variability
-#'     according to the `Scanpy` package (using the \pkg{Seurat} implementation)
-#' @param cores number of cores to use. Default is 1.
-#' @param optimizeForSpeed Boolean; when `TRUE` `COTAN` tries to use the `torch`
-#'   library to run the matrix calculations. Otherwise, or when the library is
-#'   not available will run the slower legacy code
-#' @param deviceStr On the `torch` library enforces which device to use to run
-#'   the calculations. Possible values are `"cpu"` to us the system *CPU*,
-#'   `"cuda"` to use the system *GPUs* or something like `"cuda:0"` to restrict
-#'   to a specific device
-#' @param saveObj Boolean flag; when `TRUE` saves intermediate analyses and
-#'   plots to file
-#' @param outDirCond an existing directory for the analysis output.
-#'
-#' @returns a list with a `Seurat` object along a Boolean on whether maximum
-#'   resolution has been used
-#'
-#' @importFrom Seurat CreateSeuratObject
-#' @importFrom Seurat NormalizeData
-#' @importFrom Seurat FindVariableFeatures
-#' @importFrom Seurat ScaleData
-#' @importFrom Seurat VariableFeatures
-#' @importFrom Seurat `VariableFeatures<-`
-#' @importFrom Seurat RunPCA
-#' @importFrom Seurat FindNeighbors
-#' @importFrom Seurat FindClusters
-#' @importFrom Seurat RunUMAP
-#' @importFrom Seurat DimPlot
-#'
-#' @importFrom grDevices pdf
-#' @importFrom grDevices dev.off
-#' @importFrom grDevices dev.cur
-#'
-#' @importFrom ggplot2 annotate
-#'
-#' @importFrom withr with_options
-#'
-#' @noRd
-#'
-seuratClustering <- function(rawData, cond, iter,
-                             initialResolution, minNumClusters, genesSel,
-                             cores, optimizeForSpeed, deviceStr,
-                             saveObj, outDirCond) {
-  tryCatch({
-    logThis("Creating Seurat object: START", logLevel = 2L)
-
-    srat <- CreateSeuratObject(counts = rawData,
-                               project = paste0(cond, "_reclustering_", iter),
-                               min.cells = if (iter == 1L) 3L else 1L,
-                               min.features = if (iter == 1L) 4L else 2L)
-    srat <- NormalizeData(srat)
-
-    numFeatures <- min(2000L, nrow(rawData))
-    if (str_equal(genesSel, "HGDI", ignore_case = TRUE)) {
-      obj <- COTAN(rawData)
-      obj <- proceedToCoex(obj, calcCoex = TRUE, cores = cores,
-                           optimizeForSpeed = optimizeForSpeed,
-                           deviceStr = deviceStr,
-                           saveObj = FALSE, outDir = outDirCond)
-      gdi <- getColumnFromDF(calculateGDI(obj, statType = "S",
-                                          rowsFraction = 0.05), "GDI")
-      VariableFeatures(object = srat) <-
-        names(gdi)[order(gdi, decreasing = TRUE)][seq_len(numFeatures)]
-      rm(gdi, obj)
-    } else if (str_equal(genesSel, "HVG_Seurat", ignore_case = TRUE)) {
-      srat <- FindVariableFeatures(srat, nfeatures = numFeatures,
-                                   selection.method = "vst")
-    } else if (str_equal(genesSel, "HVG_Scanpy", ignore_case = TRUE)) {
-      srat <- FindVariableFeatures(srat, nfeatures = numFeatures,
-                                   selection.method = "mean.var.plot")
-    } else {
-      stop("Unrecognised `genesSel` passed in: ", genesSel)
-    }
-
-    srat <- ScaleData(srat, features = rownames(srat))
-
-    maxRows <- nrow(srat@meta.data) - 1L
-    srat <- RunPCA(srat, features = VariableFeatures(object = srat),
-                   npcs = min(50L, maxRows))
-
-    srat <- FindNeighbors(srat, dims = seq_len(min(25L, maxRows)))
-
-    resolution <- initialResolution
-    resolutionStep <- 0.5
-    maxResolution <- initialResolution + 10.0 * resolutionStep
-    usedMaxResolution <- FALSE
-    repeat {
-      srat <- FindClusters(srat, resolution = resolution, algorithm = 2L)
-
-      # The next lines are necessary to make cluster smaller while
-      # the number of residual cells decrease and to stop clustering
-      # if the algorithm gives too many singletons.
-      usedMaxResolution <- (resolution + 0.1 * resolutionStep) > maxResolution
-      numClusters <- nlevels(factor(srat[["seurat_clusters", drop = TRUE]]))
-      if (numClusters > minNumClusters || usedMaxResolution) {
-        break
-      }
-
-      logThis(paste("Number of clusters is too small.",
-                    "Reclustering at resolution higher than:", resolution),
-              logLevel = 3L)
-
-      resolution <- resolution + resolutionStep
-    }
-
-    logThis(paste("Used resolution for Seurat clusterization is:", resolution),
-            logLevel = 2L)
-
-    # disable annoying warning about Seurat::RunUMAP()
-    srat <- with_options(list(Seurat.warn.umap.uwot = FALSE),
-                         RunUMAP(srat, umap.method = "uwot", metric = "cosine",
-                                 dims = 1L:min(c(50L, maxRows))))
-
-    if (isTRUE(saveObj)) tryCatch({
-        outFile <- file.path(outDirCond, paste0("pdf_umap_", iter, ".pdf"))
-        logThis(paste("Creating PDF UMAP in file: ", outFile), logLevel = 2L)
-        pdf(outFile)
-
-        if (iter == 1L) {
-          plot(DimPlot(srat, reduction = "umap", label = FALSE,
-                       group.by = "orig.ident"))
-        }
-
-        plot(DimPlot(srat, reduction = "umap", label = TRUE) +
-             annotate(geom = "text", x = 0.0, y = 30.0, color = "black",
-                      label = paste0("Cells number: ", ncol(rawData), "\n",
-                                     "Cl. resolution: ", resolution)))
-      }, error = function(err) {
-        logThis(paste("While saving seurat UMAP plot", err), logLevel = 1L)
-      }, finally = {
-        # Check for active device
-        if (dev.cur() > 1L) {
-          dev.off()
-        }
-      })
-    gc()
-
-    logThis("Creating Seurat object: DONE", logLevel = 2L)
-
-    # returned objects
-    return(list("SeuratObj" = srat, "UsedMaxResolution" = usedMaxResolution))
-  },
-  error = function(e) {
-    logThis(msg = paste("Seurat clusterization failed with", ncol(rawData),
-                        "cells with the following error:"), logLevel = 1L)
-    logThis(msg = conditionMessage(e), logLevel = 1L)
-    return(list("SeuratObj" = NULL, "UsedMaxResolution" = FALSE))
-  })
-}
-
 # --------------------- Uniform Clusters ----------------------
 
 #' @title Uniform Clusters
@@ -185,6 +12,152 @@ seuratClustering <- function(rawData, cond, iter,
 #' @name UniformClusters
 NULL
 
+
+### ------ Seurat Clustering -------
+
+#' @title Get a clusterization running the `Seurat` package
+#'
+#' @description The function uses the \pkg{Seurat} to clusterize the given
+#'   counts raw data.
+#'
+#' @details The parameter resolution is set at 0.5 initially, but in case of too
+#'   few clusters it can be raised up to 2.5.
+#'
+#' @param rawData The raw counts
+#' @param initialResolution The resolution to use at first in the
+#'   *clusterization* algorithm
+#' @param resolutionStep How much to increase the resolution to retry a
+#'   *clusterization* that does not respect the `minNumClusters`
+#' @param minNumClusters The minimum number of *clusters* expected from this
+#'   *clusterization*. In cases it is not reached, it will increase the
+#'   resolution of the *clusterization*
+#' @param useCoexEigen Boolean to determine whether to project the data `matrix`
+#'   onto the first eigenvectors of the **COEX** `matrix` or instead restrict
+#'   the data `matrix` to the selected genes before applying the `PCA` reduction
+#' @param dataMethod selects the method to use to create the `data.frame` to
+#'   pass to the [UMAPPlot()]. See [getDataMatrix()] for more details.
+#' @param genesSel Decides whether and how to perform the gene-selection
+#'   (defaults to `"HVG_Seurat"`). See [getSelectedGenes()] for more details.
+#' @param numGenes the number of genes to select using the above method. Will be
+#'   ignored when an explicit list of genes has been passed in
+#' @param numReducedComp the number of calculated **RDM** components
+#' @param cores number of cores to use. Default is 1.
+#' @param chunkSize number of elements to solve in batch in a single core.
+#'   Default is 1024.
+#'
+#' @returns a list with:
+#'   * `"SeuratClusters"` a `Seurat` *clusterization*
+#'   * `"CellsRDM"` the Reduced Data Matrix
+#'   * `"Resolution"` the used resolution
+#'   * `"UsedMaxResolution"` whether the maximum resolution has been reached
+#'
+#' @importFrom Seurat CreateSeuratObject
+#' @importFrom Seurat CreateDimReducObject
+#' @importFrom Seurat FindNeighbors
+#' @importFrom Seurat FindClusters
+#' @importFrom Seurat Idents
+#'
+#' @importFrom assertthat assert_that
+#'
+#' @noRd
+#'
+
+seuratClustering <- function(objCOTAN,
+                             initialResolution = 0.8,
+                             resolutionStep = 0.5,
+                             minNumClusters = 1L,
+                             useCoexEigen, dataMethod,
+                             genesSel, numGenes,
+                             numReducedComp,
+                             cores = 1L, chunkSize = 1024L) {
+  tryCatch({
+    startTime <- Sys.time()
+
+    logThis("Creating new clusterization: START", logLevel = 2L)
+
+    assert_that(numReducedComp <= getNumGenes(objCOTAN))
+
+    numReducedCompToCalc <- numReducedComp + 15L
+    cellsRDM <- calculateReducedDataMatrix(
+      objCOTAN, useCoexEigen = useCoexEigen,
+      dataMethod = dataMethod, numComp = numReducedCompToCalc,
+      genesSel = genesSel, numGenes = numGenes)
+
+    assert_that(nrow(cellsRDM) == getNumCells(objCOTAN),
+                ncol(cellsRDM) <= numReducedCompToCalc,
+                msg = "Returned PCA matrix has wrong dimensions")
+
+    # Create the Seurat object
+    srat <- CreateSeuratObject(counts = getRawData(objCOTAN),
+                               project = "clusterization")
+
+    # Add RDM manually
+    srat[["pca"]] <-
+      CreateDimReducObject(embeddings = cellsRDM,
+                           key = "PC_",
+                           assay = "RNA")
+
+    srat <- FindNeighbors(srat, dims = seq_len(numReducedComp))
+
+    resolution <- initialResolution
+    maxResolution <- initialResolution + 30.0 * resolutionStep
+
+    seuratClusters <- NULL
+    usedMaxResolution <- FALSE
+    repeat {
+      seuratClusters <- Idents(FindClusters(
+        srat,
+        resolution = resolution,
+        algorithm = 2L,      # Louvain (refined)
+        random.seed = 137    # controls igraph::cluster_louvain()
+      ))
+
+      # The next lines are necessary to make cluster smaller while
+      # the number of residual cells decrease and to stop clustering
+      # if the algorithm has gone for too long
+      usedMaxResolution <- (resolution + 0.1 * resolutionStep) > maxResolution
+      if (nlevels(seuratClusters) >= minNumClusters || usedMaxResolution) {
+        break
+      }
+
+      logThis(paste("Number of clusters is too small.",
+                    "Reclustering at resolution higher than:", resolution),
+              logLevel = 3L)
+
+      resolution <- resolution + resolutionStep
+    }
+
+    logThis(paste("Used resolution for Seurat clusterization is:", resolution),
+            logLevel = 2L)
+
+    endTime <- Sys.time()
+
+    logThis(paste("Total calculations elapsed time:",
+                  difftime(endTime, startTime, units = "secs")),
+            logLevel = 2L)
+
+    logThis("Creating new clusterization: DONE", logLevel = 2L)
+
+    rm(srat)
+    gc()
+
+    # returned objects
+    return(list("SeuratClusters" = seuratClusters, "CellsRDM" = cellsRDM,
+                "Resolution" = resolution,
+                "UsedMaxResolution" = usedMaxResolution))
+  },
+  error = function(e) {
+    logThis(msg = paste("Clusterization failed with",
+                        getNumCells(objCOTAN),
+                        "cells with the following error:"), logLevel = 1L)
+    logThis(msg = conditionMessage(e), logLevel = 1L)
+    return(list("SeuratClusters" = NULL, "CellsRDM" = NULL,
+                "Resolution" = NaN, "UsedMaxResolution" = FALSE))
+  })
+}
+
+## -------- Cells Uniform Clustering --------
+
 #' @details `cellsUniformClustering()` finds a **Uniform** *clusterizations* by
 #'   means of the `GDI`. Once a preliminary *clusterization* is obtained from
 #'   the `Seurat-package` methods, each *cluster* is checked for **uniformity**
@@ -219,14 +192,16 @@ NULL
 #' @param distance type of distance to use. Default is `"cosine"` for *DEA* and
 #'   `"euclidean"` for *Zero-One*. Can be chosen among those supported by
 #'   [parallelDist::parDist()]
-#' @param genesSel Decides whether and how to perform the gene-selection. used
-#'   for the clustering. It is a string indicating one of the following
-#'   selection methods:
-#'   * `"HGDI"` Will pick-up the genes with highest **GDI**
-#'   * `"HVG_Seurat"` Will pick-up the genes with the highest variability
-#'     via the \pkg{Seurat} package (the default method)
-#'   * `"HVG_Scanpy"` Will pick-up the genes with the highest variability
-#'     according to the `Scanpy` package (using the \pkg{Seurat} implementation)
+#' @param useCoexEigen Boolean to determine whether to project the data `matrix`
+#'   onto the first eigenvectors of the **COEX** `matrix` or instead restrict
+#'   the data `matrix` to the selected genes before applying the `PCA` reduction
+#' @param dataMethod selects the method to use to create the `data.frame` to
+#'   pass to the [UMAPPlot()]. See [getDataMatrix()] for more details.
+#' @param genesSel Decides whether and how to perform the gene-selection
+#'   (defaults to `"HVG_Seurat"`). See [getSelectedGenes()] for more details.
+#' @param numGenes the number of genes to select using the above method. Will be
+#'   ignored when an explicit list of genes has been passed in
+#' @param numReducedComp the number of calculated **RDM** components
 #' @param hclustMethod It defaults is `"ward.D2"` but can be any of the methods
 #'   defined by the [stats::hclust()] function.
 #' @param initialClusters an existing *clusterization* to use as starting point:
@@ -249,6 +224,7 @@ NULL
 #' @importFrom rlang set_names
 #' @importFrom rlang is_null
 #'
+#' @importFrom stringr str_equal
 #' @importFrom stringr str_detect
 #' @importFrom stringr str_pad
 #'
@@ -262,6 +238,14 @@ NULL
 #' @importFrom methods new
 #' @importFrom methods validObject
 #'
+#' @importFrom grDevices pdf
+#' @importFrom grDevices dev.off
+#' @importFrom grDevices dev.cur
+#'
+#' @importFrom ggplot2 annotate
+#'
+#' @importFrom withr with_options
+#'
 #' @rdname UniformClusters
 #'
 
@@ -275,17 +259,23 @@ cellsUniformClustering <- function(objCOTAN,
                                    deviceStr = "cuda",
                                    useDEA = TRUE,
                                    distance = NULL,
+                                   useCoexEigen = FALSE,
+                                   dataMethod = "",
                                    genesSel = "HVG_Seurat",
+                                   numGenes = 2000L,
+                                   numReducedComp = 25L,
                                    hclustMethod = "ward.D2",
                                    initialClusters = NULL,
                                    initialIteration = 1L,
                                    saveObj = TRUE,
                                    outDir = ".") {
+  startTime <- Sys.time()
+
   logThis("Creating cells' uniform clustering: START", logLevel = 2L)
 
   assert_that(estimatorsAreReady(objCOTAN),
-              msg = paste("Estimators lambda, nu, dispersion are not ready:",
-                          "Use proceedToCoex() to prepare them"))
+              msg = paste("Estimators `lambda`, `nu`, `dispersion` are not",
+                          "ready: Use proceedToCoex() to prepare them"))
 
   cond <- getMetadataElement(objCOTAN, datasetTags()[["cond"]])
 
@@ -299,8 +289,6 @@ cellsUniformClustering <- function(objCOTAN,
     dir.create(splitOutDir)
   }
 
-  saveSeuratObj <- saveObj && FALSE
-
   outputClusters <- set_names(rep(NA, length = getNumCells(objCOTAN)),
                               getCells(objCOTAN))
 
@@ -310,6 +298,7 @@ cellsUniformClustering <- function(objCOTAN,
 
   iter <- initialIteration - 1L
   iterReset <- -1L
+  resolutionStep <- 0.5
   numClustersToRecluster <- 0L
   srat <- NULL
   allCheckResults <- list()
@@ -326,45 +315,84 @@ cellsUniformClustering <- function(objCOTAN,
                             "a legacy `GDIThreshold` must be given"))
   }
 
+  if (isEmptyName(dataMethod)) {
+    dataMethod = "LogNormalized"
+  }
+
   repeat {
     iter <- iter + 1L
+    startLoopTime <- Sys.time()
+
+    unassignedCells <- is.na(outputClusters)
 
     logThis(paste0("In iteration ", iter, " "), logLevel = 1L, appendLF = FALSE)
     logThis(paste("the number of cells to re-cluster is",
-                  sum(is.na(outputClusters)), "cells belonging to",
+                  sum(unassignedCells), "cells belonging to",
                   numClustersToRecluster, "clusters"), logLevel = 2L)
 
+    # create COTAN sub-object
+    cellsToDrop <- getCells(objCOTAN)[!is.na(outputClusters)]
+    subObj <- dropGenesCells(objCOTAN, cells = cellsToDrop)
+
+    if ((str_equal(genesSel, "HGDI") || isTRUE(useCoexEigen)) &&
+      !isCoexAvailable(subObj)) {
+      subObj <-
+        proceedToCoex(subObj, calcCoex = TRUE, cores = cores,
+                      optimizeForSpeed = optimizeForSpeed,
+                      deviceStr = deviceStr,
+                      saveObj = FALSE, outDir = outDirCond)
+    }
+
     #Step 1
     minNumClusters <- floor(1.2 * numClustersToRecluster) + 1L
-    c(objSeurat, usedMaxResolution) %<-%
-      seuratClustering(rawData = getRawData(objCOTAN)[, is.na(outputClusters)],
-                       cond = cond, iter = iter,
+    c(testClusters, cellsRDM, resolution, usedMaxResolution) %<-%
+      seuratClustering(subObj,
                        initialResolution = initialResolution,
-                       minNumClusters = minNumClusters, genesSel = genesSel,
-                       cores = cores, optimizeForSpeed = optimizeForSpeed,
-                       deviceStr = deviceStr,
-                       saveObj = saveObj, outDirCond = splitOutDir)
+                       resolutionStep = resolutionStep,
+                       minNumClusters = minNumClusters,
+                       useCoexEigen = useCoexEigen,
+                       dataMethod = dataMethod,
+                       numReducedComp = numReducedComp,
+                       genesSel = genesSel, numGenes = numGenes,
+                       cores = cores)
 
-    if (is_null(objSeurat)) {
+    if (is_null(testClusters)) {
       logThis(paste("NO new possible uniform clusters!",
-                    "Unclustered cell left:", sum(is.na(outputClusters))),
+                    "Unclustered cell left:", sum(unassignedCells)),
               logLevel = 1L)
       break
     }
 
-    if (saveSeuratObj && iter == 1L) tryCatch({
-      # save the Seurat object to file to be reloaded later
-      saveRDS(objSeurat,
-              file.path(outDirCond, "Seurat_obj_with_cotan_clusters.RDS"))
-      },
-      error = function(err) {
-        logThis(paste("While saving seurat object", err), logLevel = 1L)
-      })
+    if (iter == initialIteration && !is_null(initialClusters)) {
+      logThis("Using passed in clusterization", logLevel = 3L)
+      testClusters <- asClusterization(initialClusters, getCells(objCOTAN))
+    }
 
-    metaData <- objSeurat@meta.data
+    ## print umap graph
+    if (isTRUE(saveObj)) tryCatch({
+      outFile <- file.path(splitOutDir, paste0("pdf_umap_", iter, ".pdf"))
+      logThis(paste("Creating PDF UMAP in file: ", outFile), logLevel = 2L)
+      pdf(outFile)
 
-    rm(objSeurat)
-    gc()
+      allCondNames <- getAllConditions(subObj)
+      condName <- ifelse(length(allCondNames) == 0L, "",
+                         allCondNames[[length(allCondNames)]])
+      title <- paste0("Cells number: ", nrow(cellsRDM))
+      plot(UMAPPlot(dataIn = cellsRDM, title = title,
+                    clusters = getCondition(subObj, condName)))
+
+      title <- paste0(title, "\nCl. resolution: ", resolution,
+                      ifelse(isTRUE(usedMaxResolution), " [max]", ""))
+      plot(UMAPPlot(dataIn = cellsRDM, title = title,
+                    clusters = testClusters))
+    }, error = function(err) {
+      logThis(paste("While saving seurat UMAP plot", err), logLevel = 1L)
+    }, finally = {
+      # Check for active device
+      if (dev.cur() > 1L) {
+        dev.off()
+      }
+    })
 
     # Step 2
 
@@ -374,76 +402,76 @@ cellsUniformClustering <- function(objCOTAN,
     numClustersToRecluster <- 0L
     cellsToRecluster <- vector(mode = "character")
 
-    testClusters <- factor(metaData[["seurat_clusters"]])
-    allCells <- rownames(metaData)
-    names(testClusters) <- allCells
-    if (iter == initialIteration && !is_null(initialClusters)) {
-      logThis("Using passed in clusterization", logLevel = 3L)
-      testClusters <- asClusterization(initialClusters, getCells(objCOTAN))
-      allCells <- names(testClusters)
-    }
+    allCells <- names(testClusters)
     testClList <- toClustersList(testClusters)
 
     globalClName <- ""
 
-    for (clName in names(testClList)) {
-      logThis("*", logLevel = 1L, appendLF = FALSE)
-      logThis(paste0(" checking uniformity of cluster '", clName,
-                     "' of ", length(testClList), " clusters"),
-              logLevel = 2L)
+    minimumUTClusterSize <- 20L
+    maxClusterSize <- max(lengths(testClList))
 
-      if (clName == "singleton") {
-        next
-      }
+    if (maxClusterSize >= minimumUTClusterSize) {
+      for (clName in names(testClList)) {
+        logThis("*", logLevel = 1L, appendLF = FALSE)
+        logThis(paste0(" checking uniformity of cluster '", clName,
+                       "' of ", length(testClList), " clusters"),
+                logLevel = 2L)
 
-      globalClName <-
-        paste0(str_pad(iter, width = 2L, pad = "0"), "_",
-               str_pad(clName, width = 4L, pad = "0"))
+        globalClName <-
+          paste0(str_pad(iter, width = 2L, pad = "0"), "_",
+                 str_pad(clName, width = 4L, pad = "0"))
 
-      cells <- testClList[[clName]]
-      if (length(cells) < 20L) {
-        logThis(paste("cluster", globalClName, "has too few cells:",
-                      "will be reclustered!"), logLevel = 2L)
-
-        numClustersToRecluster <- numClustersToRecluster + 1L
-        cellsToRecluster <- c(cellsToRecluster, cells)
-      } else {
-        checkResults <- tryCatch(
-          checkClusterUniformity(objCOTAN = objCOTAN,
-                                 clusterName = globalClName,
-                                 cells = cells,
-                                 checker = checker,
-                                 cores = cores,
-                                 optimizeForSpeed = optimizeForSpeed,
-                                 deviceStr = deviceStr,
-                                 saveObj = saveObj,
-                                 outDir = splitOutDir),
-          error = function(err) {
-            logThis(paste("while checking cluster uniformity", err),
-                    logLevel = 0L)
-            logThis("marking cluster as not uniform", logLevel = 2L)
-            return(checker)
-          })
-
-        invisible(validObject(checkResults))
-
-        allCheckResults <- append(allCheckResults, checkResults)
-        names(allCheckResults)[length(allCheckResults)] <- globalClName
-
-        if (!checkResults@isUniform) {
-          logThis(paste("cluster", globalClName, "has too high GDI:",
+        cells <- testClList[[clName]]
+        if (length(cells) < 20L) {
+          logThis(paste("cluster", globalClName, "has too few cells:",
                         "will be reclustered!"), logLevel = 2L)
 
           numClustersToRecluster <- numClustersToRecluster + 1L
           cellsToRecluster <- c(cellsToRecluster, cells)
         } else {
-          logThis(paste("cluster", globalClName, "is uniform"), logLevel = 2L)
-        }
-        logThis("", logLevel = 2L)
+          checkResults <- tryCatch(
+            checkClusterUniformity(objCOTAN = objCOTAN,
+                                   clusterName = globalClName,
+                                   cells = cells,
+                                   checker = checker,
+                                   cores = cores,
+                                   optimizeForSpeed = optimizeForSpeed,
+                                   deviceStr = deviceStr,
+                                   saveObj = saveObj,
+                                   outDir = splitOutDir),
+            error = function(err) {
+              logThis(paste("while checking cluster uniformity", err),
+                      logLevel = 0L)
+              logThis("marking cluster as not uniform", logLevel = 2L)
+              return(checker)
+            })
 
-        rm(checkResults)
-        gc()
+          invisible(validObject(checkResults))
+
+          allCheckResults <- append(allCheckResults, checkResults)
+          names(allCheckResults)[length(allCheckResults)] <- globalClName
+
+          if (!checkResults@isUniform) {
+            logThis(paste("cluster", globalClName, "has too high GDI:",
+                          "will be reclustered!"), logLevel = 2L)
+
+            numClustersToRecluster <- numClustersToRecluster + 1L
+            cellsToRecluster <- c(cellsToRecluster, cells)
+          } else {
+            logThis(paste("cluster", globalClName, "is uniform"), logLevel = 2L)
+          }
+          logThis("", logLevel = 2L)
+
+          rm(checkResults)
+          gc()
+        }
       }
+    } else {
+      # all clusters are too small: nothing to do
+      logThis(paste("All clusters in iteration ", iter, "have too few cells"),
+              logLevel = 2L)
+      numClustersToRecluster <- length(testClList)
+      cellsToRecluster <- allCells
     }
 
     logThis("", logLevel = 1L)
@@ -455,8 +483,11 @@ cellsUniformClustering <- function(objCOTAN,
       warning("In iteration '", iter, "' no uniform clusters found!")
       # Another iteration can be attempted as the minimum number of clusters
       # will be higher. This happens unless the resolution already reached
-      # its maximum. In the latter case we simply stop here.
-      if (isTRUE(usedMaxResolution)) {
+      # its maximum. In the latter case we simply stop here, unless some UT
+      # cluster was found. In the latter case we restart from the minimal
+      # resolution, but with larger resolution steps and try to re-cluster with
+      # no minimal number of clusters
+      if (isTRUE(usedMaxResolution) || maxClusterSize < minimumUTClusterSize) {
         logThis("Max resolution reached", logLevel = 1L)
         if (iterReset != -1L) {
           logThis("Cannot clusterize anything more", logLevel = 2L)
@@ -467,6 +498,9 @@ cellsUniformClustering <- function(objCOTAN,
                   logLevel = 2L)
           numClustersToRecluster <- 0L
           iterReset <- iter
+          if (maxClusterSize >= minimumUTClusterSize) {
+            resolutionStep <- resolutionStep + 0.3
+          }
         }
       }
     } else {
@@ -501,14 +535,25 @@ cellsUniformClustering <- function(objCOTAN,
       break
     }
 
-    if (length(cellsToRecluster) < 40L
-        || iter > maxIterations + initialIteration) {
-      logThis(paste("NO new possible uniform clusters!",
-                    "Unclustered cell left:", length(cellsToRecluster)),
+    if (length(cellsToRecluster) < 40L ||
+        iter >= maxIterations + initialIteration - 1L) {
+      logThis(ifelse(length(cellsToRecluster) < 40L,
+                     "Stopping: too few cells left to cluster",
+                     "Stopping: maximum number of iterations reached"),
               logLevel = 1L)
+      logThis(paste("Unclustered cell left:", length(cellsToRecluster)),
+              logLevel = 2L)
       break
     }
+
     rm(cellsToRecluster)
+    gc()
+
+    endLoopTime <- Sys.time()
+
+    logThis(paste("Loop calculations elapsed time:",
+                  difftime(endLoopTime, startLoopTime, units = "secs")),
+            logLevel = 2L)
   } # End repeat
 
   logThis(paste("The final raw clusterization contains [",
@@ -557,54 +602,26 @@ cellsUniformClustering <- function(objCOTAN,
     })
   names(allCheckResults) <- permMap[names(allCheckResults)]
 
-  outputList <- list("clusters" = factor(outputClusters), "coex" = outputCoexDF)
-
   if (isTRUE(saveObj)) tryCatch({
       outFile <- file.path(outDirCond, "split_check_results.csv")
       write.csv(checkersToDF(allCheckResults), file = outFile, na = "NaN")
 
-      if (saveSeuratObj) {
-        clusterizationName <-
-          paste0(as.roman(length(getClusterizations(objCOTAN)) + 1L))
-
-        srat <-
-          readRDS(file.path(outDirCond, "Seurat_obj_with_cotan_clusters.RDS"))
-
-        if (!setequal(rownames(srat@meta.data), names(outputClusters))) {
-          warning("List of cells got corrupted")
-          return(outputList)
-        }
-
-        srat@meta.data <-
-          setColumnInDF(srat@meta.data,
-                        colToSet = outputClusters[rownames(srat@meta.data)],
-                        colName = paste0("COTAN_", clusterizationName))
-
-        if (!dim(srat)[[2L]] == getNumCells(objCOTAN)) {
-          warning("Number of cells got wrong")
-          return(outputList)
-        }
-
-        logThis("Cluster, UMAP and Saving the Seurat dataset", logLevel = 2L)
-
-        srat <- FindNeighbors(srat, dims = 1L:25L)
-        srat <- FindClusters(srat, resolution = 0.5, algorithm = 2L)
-        srat <- RunUMAP(srat, umap.method = "uwot",
-                        metric = "cosine", dims = 1L:25L)
-
-        saveRDS(srat, file.path(outDirCond,
-                                "Seurat_obj_with_cotan_clusters.RDS"))
-      }
+      outFile <- file.path(outDirCond, "split_clusterization.csv")
+      write.csv(as.data.frame(list("split" = outputClusters)),
+                file = outFile, na = "NaN")
     },
     error = function(err) {
-      logThis(paste("While saving seurat object", err), logLevel = 1L)
+      logThis(paste("While saving results csv", err), logLevel = 1L)
     }
   )
 
-  rm(srat)
-  gc()
+  endTime <- Sys.time()
+
+  logThis(paste("Total calculations elapsed time:",
+                difftime(endTime, startTime, units = "secs")),
+          logLevel = 2L)
 
   logThis("Creating cells' uniform clustering: DONE", logLevel = 2L)
 
-  return(outputList)
+  return(list("clusters" = outputClusters, "coex" = outputCoexDF))
 }
